# -*- coding: utf-8 -*-
import logging
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import TimeoutException
from .drivers_tools import web_drivers_tuple
from .finders import find_element


def is_field_exist(driver=None, field=None, until=5):
    """
    Test if the field given as a {"type":"id","value":"toto"} dictionary exists.
    :param driver: a selenium web driver
    :param field: a dictionary
    :param until: an int as the wait time
    :raise TypeError: driver isn't of the expected type
    :return: a web element if exist, None otherwise
    """
    try:
        if driver is None or not isinstance(driver, web_drivers_tuple()):
            raise TypeError("Driver is expected as a WebDriver")

        switcher = {
            "id": By.ID,
            "name": By.NAME,
            "class_name": By.CLASS_NAME,
            "css": By.CSS_SELECTOR,
            "link_text": By.LINK_TEXT,
            "partial_link_text": By.PARTIAL_LINK_TEXT,
            "tag_name": By.TAG_NAME,
            "xpath": By.XPATH
        }

        return WebDriverWait(driver, until).until(
            EC.presence_of_element_located((switcher[field["type"]], field["value"])))
    except TypeError as assertion:
        logging.error("information.is_field_exist raised an assertion with following"
                      f" input driver:'{driver}', field:'{field}' and until:'{until}'. "
                      f"Assertion is '{assertion.args}'")
        raise
    except TimeoutException:
        logging.warning(f"information.is_field_exist raised a TimeoutException for "
                        f"the following field '{field}'")
        return None


def is_field_contains_text(driver=None, field=None, text=None):
    """
    Check if the given field contains the text either as a DOM text or value text.
    {"type":"id","value":"toto"}
    :param driver: a selenium web driver
    :param field: a dictionary
    :param text: a string for the text to be in the field
    :raise AssertionError: driver isn't of the expected type
    :return: True if the field contains the text, False otherwise
    """
    try:
        assert driver is not None and isinstance(driver, web_drivers_tuple()),\
            "Driver is expected."
        element = is_field_exist(driver=driver, field=field)
        return (element.text is not None and text in element.text) or (
                element.get_attribute("value") is not None
                and text in element.get_attribute("value"))
    except AssertionError as assertion:
        logging.error("information.is_field_exist raised an assertion with "
                      "following input driver:'{}', field:'{}' and text:'{}'. "
                      "Assertion is '{}'".format(driver, field, text, assertion.args))
        raise


def is_alert_present(driver=None, until=5):
    """
    Tells if whatever an alert is present or not.
    :param driver: a selenium web driver
    :param until: time to wait.
    :return: True or False.
    """
    try:
        assert driver is not None and isinstance(driver, web_drivers_tuple()), "Driver is expected."
        if WebDriverWait(driver, until).until(EC.alert_is_present()):
            return True
    except AssertionError as assertion:
        logging.error("information.is_alert-present raised an assertion with "
                      "following input driver:'{}' and until:'{}'. "
                      "Assertion is '{}'".format(driver, until, assertion.args))
        raise
    except TimeoutException:
        return False


def element_text(driver=None, field=None):
    """
    Return the text of the element
    :param driver: a selenium web driver
    :param field: a dictionary
    :raise AssertionError: from eaifinders.find_element method
    :raise Exception: if text and value are defined but not identical
    :return: the element text or value, empty if no text or value
    """
    element = find_element(driver=driver, field=field)
    element_text_ = element.text
    element_value = element.get_attribute("value")

    if element_text_ and not element_value:
        return element_text_
    elif not element_text_ and element_value:
        return element_value
    elif element_text_ is not None and element_value is not None \
            and element_text_ in element_value and element_value in element_text_:
        return element_text_
    elif element_text_ is not None and element_value is not None and (
            element_text_ not in element_value or element_value not in element_text_):
        raise Exception("Can't serve the element 'text' having both data for text and"
                        " attribute value")
    else:
        return ""


def how_many_windows(driver=None):
    """
    Return the number of windows at the method execution.
    :param driver: a selenium web driver
    :return: the number of windows
    """
    try:
        assert driver is not None and isinstance(driver, web_drivers_tuple()),\
            "Driver is expected."
        return len(driver.window_handles)
    except AssertionError as assertion:
        logging.error("information.is_alert-present raised an assertion with "
                      "following input driver:'{}'. "
                      "Assertion is '{}'".format(driver, assertion.args))
        raise


def is_field_displayed(driver=None, field=None):
    """
    Check if the element is displayed. You may not interact with it.
    :param driver: a selenium web driver
    :param field: a dictionary
    :return: Boolean. True if element is displayed, false otherwise.
    """
    element = is_field_exist(driver=driver, field=field)
    if element is not None:
        return element.is_displayed()
    else:
        return False


def is_field_enabled(driver=None, field=None, attribute=None):
    """
    Check if the element is enabled.
    :param attribute: the attribute to check for enabled
    :param driver: a selenium web driver
    :param field: a dictionary
    :return: Boolean. True if element is enabled, false otherwise for non attribute.
            Attribute string otherwise
    """
    element = is_field_exist(driver=driver, field=field)
    if element is not None:
        if attribute is not None:
            return element.get_attribute(attribute)
        else:
            return element.is_enabled()
    else:
        return False


def where_am_i(driver=None):
    """
    Return the current location URL
    :param driver: a selenium web driver
    :return: String. The current URL
    """
    if driver is None or not isinstance(driver, web_drivers_tuple()):
        raise TypeError("Driver is expected as a WebDriver")

    return driver.current_url


def is_checkbox_checked(driver=None, field=None, is_angular=False):
    """
    Return the checkbox status
    :param driver: a selenium web driver
    :param field: a dictionary
    :param is_angular: boolean
    :return: boolean
    """
    element = is_field_exist(driver=driver, field=field)

    if is_angular:
        return bool(element.get_attribute("ng-reflect-checked"))
    else:
        return bool(element.get_attribute("checked"))


def retrieve_tabular(driver=None, field=None, row_and_col=("tr", "td", "th")) -> list:
    """
    Return the tabular as a list of elements.
    Elements are either lists or dictionaries depending on the presence of headers
    :param driver: a selenium web driver
    :param row_and_col: a tuple of row and col tags plus header tag. Defaulted to ('tr', 'td', 'th').
     Please mind the order. ROW, COL, COL HEADER
    :param field: a dictionary
    :return: list
    """
    tabular = find_element(driver, field)
    rows = tabular.find_elements_by_tag_name(row_and_col[0])
    tabular_as_list = list()

    for row in rows:
        columns = row.find_elements_by_tag_name(row_and_col[1])
        if not columns:
            columns = row.find_elements_by_tag_name(row_and_col[2])
        tabular_as_list.append([column.text for column in columns])

    return tabular_as_list



# Could held the webdriver instance so that we could interact with the browser.
# -*- coding: utf-8 -*-
import os.path
import logging
import tempfile
from deprecated import deprecated
from shutil import rmtree
from datetime import datetime
from selenium import webdriver
# Driver's options
from selenium.webdriver.chrome.options import Options as ChrOptions
from selenium.webdriver.firefox.options import Options as FfOptions
from selenium.webdriver.edge.options import Options as EdgOptions
from selenium.webdriver.opera.options import Options as OpeOptions
from selenium.webdriver.remote.webelement import WebElement
# Driver Manager
from webdriver_manager.chrome import ChromeDriverManager
from webdriver_manager.firefox import GeckoDriverManager
from webdriver_manager.microsoft import EdgeChromiumDriverManager
from webdriver_manager.opera import OperaDriverManager
# Self 
from .navigators import go_to_url, enter_frame, go_to_window
from .finders import find_element, find_elements, find_from_elements, \
    find_sub_element_from_element
from .actions import fill_element, fill_elements, select_in_dropdown, set_checkbox, \
    click_element, select_in_angular_dropdown, hover_element, select_in_elements
from .alerts import alert_message, intercept_alert
from .information import is_alert_present, is_field_exist, is_field_contains_text, \
    element_text, is_field_displayed, is_field_enabled, how_many_windows, where_am_i, \
    is_checkbox_checked, retrieve_tabular
from .drivers_tools import fullpage_screenshot, move_to

log = logging.getLogger(__name__)


class BrowserServer:
    """
    The BrowserServer is a convenient object which instantiate a selenium web driver.
    It serves all automaton tools such as finders, actions, alert and navigators.
    Although these tools can be used separately and are tested separately,
    it will be easier to use them from this class.
    Please see the eaiautomatontools specific doctest.
    Todo set others webdriver executables
    """

    __OPTIONS_SWITCHER = {
        "chrome": ChrOptions,
        "headless-chrome": ChrOptions,
        "firefox": FfOptions,
        "opera": OpeOptions,
        "edge": EdgOptions
    }
    
    __DRIVER_MANAGER = {
            "chrome": ChromeDriverManager,
            "headless-chrome": ChromeDriverManager,
            "firefox": GeckoDriverManager,
            "edge": EdgeChromiumDriverManager,
            "opera": OperaDriverManager
        }
    
    __WEB_DRIVERS = {
            "chrome": webdriver.Chrome,
            "headless-chrome": webdriver.Chrome,
            "firefox": webdriver.Firefox,
            "edge": webdriver.Edge,
            "safari": webdriver.Safari,
            "opera": webdriver.Opera
        }
    
    def __init__(self):
        # Definition of private attributes and references
        # All strings allowed for the browser name and version
        self.__authorized_name_version = ("chrome", "firefox", "opera",
                                          "edge", "safari", "headless-chrome")
        # Browser window's status
        self.__launched = False

        # Screenshot save folder
        self.__temp_save_to = os.path.normpath(os.path.join(tempfile.gettempdir(),
                                                            "automaton_screenshots"))
        if os.path.exists(self.__temp_save_to):
            rmtree(self.__temp_save_to)
        os.makedirs(self.__temp_save_to)
        log.debug(self.__temp_save_to)

        # Definition of private attributes
        self.__web_driver = None  # The webdriver
        self.__browser_name = None  # The browser's name used
        self.__driver_path = None  # The path to the webdriver executable

        # Definition of options
        self.__driver_options = []

    @property
    def driver_options(self):
        return self.__driver_options

    @driver_options.setter
    def driver_options(self, options):
        if isinstance(options, list):
            if not options:
                self.__driver_options = []
            elif all(isinstance(item, str) for item in options):
                self.__driver_options.extend(options)
            else:
                log.warning(f"Options '{options}' is not of the proper type")
        elif isinstance(options, str):
            self.__driver_options.append(options)
        else:
            log.warning(f"Options '{options}' is not of the proper type")

    @property
    def is_launched(self):
        return self.__launched

    @property
    def webdriver(self):
        """Read only webdriver"""
        return self.__web_driver

    @property
    def browser_name(self):
        """Read - Set browser name with restrictions"""
        return self.__browser_name

    @browser_name.setter
    def browser_name(self, name: str):
        if name.casefold() not in self.__authorized_name_version:
            raise ValueError(f"Unknown browser name. "
                             f"Get {name.casefold()} instead of {self.__authorized_name_version}")
        self.__browser_name = name.casefold()
        self.__driver_path = None

    @property
    def driver_path(self):
        return self.__driver_path

    @driver_path.setter
    def driver_path(self, new_path: str):
        if new_path is not None and new_path:
            if os.path.exists(new_path):
                self.__driver_path = new_path
            else:
                raise ValueError("Expecting an existing file")
        else:
            raise ValueError("Expecting a non empty path")

    @staticmethod
    def __serve_time():
        return int(datetime.now().timestamp() * 1000000)

    def serve(self):
        """
        Create the webdriver instance
        Todo incorporate the outside module resources locator
        TODO Handle the case where you don't want to use the webdriver manager
        :return: 0 if success
        """
        # todo use the data in order to launch the expected webdriver
        if self.browser_name == "safari":
            self.__web_driver = BrowserServer.__WEB_DRIVERS[self.browser_name](
                executable_path=self.driver_path)
        elif self.browser_name == "chrome":
            # Hack https://stackoverflow.com/questions/64927909/
            # failed-to-read-descriptor-from-node-connection-a-device-attached-to-the-system
            if self.__driver_path is None:
                self.__driver_path = BrowserServer.__DRIVER_MANAGER[self.browser_name]().install() 
            option = BrowserServer.__OPTIONS_SWITCHER[self.browser_name]()
            option.add_experimental_option('excludeSwitches', ['enable-logging'])
            if self.driver_options:
                for opt in self.driver_options:
                    option.add_argument(opt)
            self.__web_driver = BrowserServer.__WEB_DRIVERS[self.browser_name](
                    executable_path=self.__driver_path,
                    options=option
                )
        elif self.browser_name == "headless-chrome":
            option = BrowserServer.__OPTIONS_SWITCHER[self.browser_name]()
            option.add_experimental_option('excludeSwitches', ['enable-logging'])
            if self.driver_options:
                for opt in self.driver_options:
                    option.add_argument(opt)
            option.headless = True
            if self.__driver_path is None:
                self.__driver_path = BrowserServer.__DRIVER_MANAGER[self.browser_name]().install()
            self.__web_driver = BrowserServer.__WEB_DRIVERS[self.browser_name](
                    executable_path=self.__driver_path,
                    options=option
                )
        elif self.browser_name is not None:
            if self.__driver_path is None:
                self.__driver_path = BrowserServer.__DRIVER_MANAGER[self.browser_name]().install()

            option = BrowserServer.__OPTIONS_SWITCHER[self.browser_name]()
            if self.driver_options:
                for opt in self.driver_options:
                    option.add_argument(opt)
            self.__web_driver = BrowserServer.__WEB_DRIVERS[self.browser_name](
                    executable_path=self.__driver_path,
                    options=option)

        else:
            raise AttributeError("You must set a browser name. "
                                 f"Use one of '{self.__authorized_name_version}'")
        self.__launched = True
        return 0

    def close(self):
        """
        Close the webdriver
        :return:
        """
        self.webdriver.quit()
        self.__web_driver = None
        self.__launched = False
        return 0

    def full_screenshot(self, filename: str):
        return fullpage_screenshot(self.webdriver, filename)

    def take_a_screenshot(self, save_to: str = None, is_full_screen: bool = True):
        """
        Take a screenshot and save the file to the given folder
        Each screenshot is followed by a timestamp.
        :param is_full_screen:
        :param save_to:
        :return: 0 if success
        """
        try:
            # Process the filename
            if save_to is None:
                filename = os.path.join(
                    self.__temp_save_to, "screenshot-{}.png".format(self.__serve_time()))
            else:
                log.debug("Try to save to '{}'".format(save_to))
                filename = os.path.join(save_to, "screenshot-{}.png".format(self.__serve_time()))
            if is_full_screen:
                result = self.full_screenshot(filename)
            else:
                result = self.webdriver.get_screenshot_as_file(filename)
            if not result:
                log.error("The screenshot could not be done."
                          " Please check if the file path is correct."
                          " Get '{}'".format(repr(result)))
                raise IOError("The screenshot could not be done."
                              " Please check if the file path is correct."
                              " Get '{}'".format(repr(result)))
            return os.path.realpath(filename)
        except IOError as io_error:
            log.error(f"Screenshot raised an IO error '{io_error.args[0]}'")
            raise IOError(io_error.args[0]) from None
        except Exception as exception:
            log.error(f"Screenshot raised an error '{exception.args[0]}'")
            raise Exception(exception.args[0]) from None

    # Start of convenient usage of the automaton tools
    # Navigation
    def go_to(self, url=None):
        return go_to_url(driver=self.webdriver, url=url)

    def enter_frame(self, field=None):
        return enter_frame(driver=self.webdriver, field=field)

    def go_to_window(self, handle=None, title=None):
        return go_to_window(driver=self.webdriver, handle=handle, title=title)

    # Finders
    def find_element(self, field: dict = None, web_element: WebElement = None):
        """Find element using the current webdriver or the provided WebElement"""
        return find_element(driver=self.webdriver, field=field, web_element=web_element)

    def find_elements(self, field: dict = None, web_element: WebElement = None):
        """Find elements using the current webdriver or the provided WebElement"""
        return find_elements(driver=self.webdriver, field=field, web_element=web_element)

    def find_from_elements(self, field=None, text=None, web_element: WebElement = None):
        """Find element from a list of elements based on the text (exact match)"""
        return find_from_elements(driver=self.webdriver,
                                  field=field,
                                  text=text,
                                  web_element=web_element)

    @deprecated(version="1.0.5", reason="You should user find_element with a web_element")
    def find_sub_element_from_element(self, field=None):
        """
        Find the webelement contained in another webelement.
        If no parent is provided try to return the webelement
        :param field:
        :return:
        """
        if "parent" not in field:
            return self.find_element(field=field)
        element = self.find_element(field=field["parent"])
        return find_sub_element_from_element(element, field=field)

    # Actions
    def fill_element(self, field=None, value=None):
        return fill_element(driver=self.webdriver, field=field, value=value)

    def fill_elements(self, fields=None, data=None):
        return fill_elements(driver=self.webdriver, fields=fields, data=data)

    # TODO add unit test
    def click_element(self, field=None, web_element=None):
        return click_element(driver=self.webdriver,
                             field=field,
                             web_element=web_element)

    def select_in_dropdown(self, field=None, visible_text=None, value=None):
        return select_in_dropdown(driver=self.webdriver,
                                  field=field,
                                  visible_text=visible_text,
                                  value=value)

    def move_to(self, web_element: WebElement):
        return move_to(driver=self.webdriver,
                       element=web_element)

    # TODO add unit test
    def select_in_angular_dropdown(self, root_field=None, visible_text=None):
        return select_in_angular_dropdown(driver=self.webdriver,
                                          root_field=root_field,
                                          visible_text=visible_text)

    def set_checkbox(self, field=None, is_checked=None):
        return set_checkbox(driver=self.webdriver,
                            field=field,
                            is_checked=is_checked)

    # TODO add unit test
    def hover_element(self, field=None):
        return hover_element(driver=self.webdriver, field=field)

    # TODO add unit test
    def select_in_elements(self, field=None, displayed_text=None):
        return select_in_elements(driver=self.webdriver,
                                  field=field,
                                  displayed_text=displayed_text)

    # Alerts
    def alert_message(self):
        return alert_message(driver=self.webdriver)

    def intercept_alert(self, messages=None, accept=True, value=None):
        return intercept_alert(driver=self.webdriver, messages=messages,
                               accept=accept, value=value)

    # Information

    def is_alert_present(self, until=5):
        return is_alert_present(driver=self.webdriver, until=until)

    def is_field_exist(self, field=None, until=5):
        return is_field_exist(driver=self.webdriver, field=field, until=until)

    def is_field_contains_text(self, field=None, text=None):
        return is_field_contains_text(driver=self.webdriver, field=field, text=text)

    def element_text(self, field=None):
        return element_text(driver=self.webdriver, field=field)

    def is_field_displayed(self, field=None):
        return is_field_displayed(driver=self.webdriver, field=field)

    def is_field_enabled(self, field=None, attribute=None):
        return is_field_enabled(driver=self.webdriver, field=field, attribute=attribute)

    def how_many_windows(self):
        return how_many_windows(driver=self.webdriver)

    def retrieve_tabular(self, field=None, row_and_col=("tr", "td", "th")):
        return retrieve_tabular(driver=self.webdriver, field=field, row_and_col=row_and_col)

    # TODO add unit test
    def where_am_i(self):
        return where_am_i(driver=self.webdriver)

    # TODO add unit test
    def is_checkbox_checked(self, field=None, is_angular=False):
        return is_checkbox_checked(driver=self.webdriver, field=field, is_angular=is_angular)




# -*- coding: utf-8 -*-
from .finders import find_element
from .drivers_tools import web_drivers_tuple
"""
The eainavigators tool box purpose is to provide some limited but heavy used methods
in order to browse to URL and navigate from browser tabs or windows.

The behaviour of these methods are known so that it will ease the debugging.
"""

# Todo: Add a logger so that it will be easy to search for issues.


def go_to_url(driver=None, url=None):
    """
    Go to a specific url
    :param driver: a selenium web driver
    :param url: a URL as a string
    :raise AssertionError: if driver is not defined
    :return: 0 if succeed
    """
    assert driver is not None and isinstance(driver, web_drivers_tuple()), "Driver is expected."
    driver.get(url)
    return 0


def enter_frame(driver=None, field=None):
    """
    Switch to a frame given by its field identifier
    :param driver: a selenium web driver
    :param field: a dictionary like {"type":"id", "value":"myID"}
    :raise AssertionError: if driver is not defined
    :return: 0 if succeed
    """
    assert driver is not None and isinstance(driver, web_drivers_tuple()), "Driver is expected."

    driver.switch_to.frame(find_element(driver=driver, field=field))
    return 0


def go_to_window(driver=None, handle=None, title=None):
    """
    Switch to a window either giving the handle or the window's title.
    :param driver: a selenium web driver
    :param handle: a window's reference as a string
    :param title: a window's title as a string
    :raise Exception: The number of window with the title is different from 1.
    :raise AssertionError: if driver is not defined
    :return: 0 if succeed
    """
    assert driver is not None and isinstance(driver, web_drivers_tuple()), "Driver is expected."

    if handle is None and title is None:
        raise ValueError("Expect at least one value either the handle or the window title")

    if handle is not None:
        driver.switch_to.window(handle)
    if title is not None:
        current_handle = driver.current_window_handle
        handles = driver.window_handles
        search_result = []
        for handle in handles:
            driver.switch_to.window(handle)
            if driver.title == title:
                search_result.append(handle)
        if len(search_result) != 1:
            driver.switch_to.window(current_handle)
            raise Exception("Too many windows with the same title to switch to. Staying on the "
                            "current window.")
        else:
            driver.switch_to.window(search_result[0])

    return 0



# -*- coding: utf-8 -*-
import logging
from selenium.common.exceptions import ElementNotSelectableException, NoAlertPresentException, \
    ElementNotInteractableException
from .drivers_tools import web_drivers_tuple


def intercept_alert(driver=None, messages=None, accept=True, value=None):
    """
        intercept_alert goal is to provide an easy interface to handle alerts.

        Only the web driver is mandatory.

        Giving only the web driver accept the alert whatever the message is.

        Giving a list of messages, ensure that the alert message is one of the list.
    :param value: The string to be entered in the prompt alert.
    :param driver: a selenium webdriver
    :param messages: A list of string which should include the alert message.
    :param accept: boolean defaulted to True. Accept or dismiss the alert.
    :raise AssertionError: driver, messages, accept and value aren't of the expected type
    :raise ElementNotSelectableException: when trying to send keys on a non-prompt alert
    :raise NoAlertPresentException: when trying to intercept an alert which not present
    :raise Exception: when the alert message is not one of the expected
    :return: 0 if success
    """
    try:
        assert driver is not None and isinstance(driver, web_drivers_tuple()), "Driver is expected."
        assert messages is None or isinstance(messages, list), "Messages should be a list or None"
        assert isinstance(accept, bool), "Accept is boolean True or False"
        assert value is None or isinstance(value, str), "Value is None or a string"

        alert_object = driver.switch_to.alert

        if (
            messages is not None
            and isinstance(messages, list)
            and alert_object.text not in messages
        ):
            raise Exception("Message not found")

        if value is not None:
            alert_object.send_keys(value)

        if accept:
            alert_object.accept()
        else:
            alert_object.dismiss()

        return 0
    except AssertionError as assertion:
        logging.error("alerts.intercept_alert raised an assertion with following"
                      " input driver:'{}', message:'{}' and accept:'{}'."
                      " Assertion is '{}'".format(driver, messages, accept, assertion.args))
        raise
    except ElementNotSelectableException as not_selectable:
        logging.error("alerts.intercept_alert can't fill the alert popup with '{}'"
                      " as there is no input field.\nGet {}".format(value, not_selectable.args))
        raise ElementNotSelectableException(
            "Can't fill the alert popup with '{}' as there is no "
            "input field.".format(value)) from None
    except ElementNotInteractableException as not_interactable:
        logging.error(not_interactable)
        raise ElementNotInteractableException("Cannot found an input field") from None
    except NoAlertPresentException as no_alert:
        logging.error(
            "alerts.intercept_alert can't interact with an alert as there is no "
            "displayed alert.\nGet {}".format(no_alert.args))
        raise NoAlertPresentException("Can't interact with an alert as there is no "
                                      "displayed alert") from None


def alert_message(driver=None):
    """
        Return the message displayed in the alert.
    :param driver: a selenium webdriver
    :raise AssertionError: driver isn't of the expected type
    :raise NoAlertPresentException: when trying to intercept an alert which not present
    :return: 0 if success
    """
    try:
        assert driver is not None and isinstance(driver, web_drivers_tuple()), "Driver is expected."
        alert_object = driver.switch_to.alert

        return alert_object.text
    except AssertionError as assertion:
        logging.error("alerts.alert_message raised an assertion with following input"
                      " driver:'{}'. Assertion is '{}'".format(driver, assertion.args))
        raise
    except NoAlertPresentException as no_alert:
        logging.error("alerts.intercept_alert can't interact with an alert as there"
                      " is no displayed alert.\nGet {}".format(no_alert.args))
        raise NoAlertPresentException("Can't interact with an alert as there is no"
                                      " displayed alert") from None



# -*- coding: utf-8 -*-
import logging
from time import sleep

from selenium.webdriver.remote.webelement import WebElement

from .finders import find_element, find_from_elements, find_sub_element_from_element
from .information import is_field_exist, is_field_displayed
from .drivers_tools import driver_field_validation, web_drivers_tuple, web_element_validation
from selenium.common.exceptions import InvalidElementStateException, NoSuchElementException, \
    StaleElementReferenceException
from selenium.webdriver.support.select import Select
from selenium.webdriver.common.keys import Keys
from selenium.webdriver import ActionChains

log = logging.getLogger(__name__)


def fill_elements(driver=None, fields=None, data=None):
    """
    Fill a field set with data where the field and data are identified by the same key
    Keys are taken from the data set dictionary
    :param driver: a selenium web driver
    :param fields: a dictionary
    :param data: a dictionary
    :raise AssertionError: driver is not define, fields and data doesn't have the same keys
    :raise InvalidElementStateException: rethrown from fill_element
    :raise Exception: all other issues
    :return: 0 if success
    """
    try:
        assert all(key in fields.keys() for key in
                   data.keys()), \
            "Missing fields for the given data. Data keys '{}'. Fields keys '{}'".format(
                data.keys(),
                fields.keys())  # Check that the fields dictionary contains enough keys

        for key in data:
            fill_element(driver=driver, field=fields[key], value=data[key])

        return 0
    except AssertionError as assertion:
        log.error("actions.fill_elements raised an assertion with following"
                  " input driver:'{}', fields:'{}' and data:'{}'."
                  " Assertion is '{}'".format(driver, fields, data, assertion.args))
        raise
    except InvalidElementStateException:
        raise
    except Exception as exception:
        logging.error(exception.args)
        raise Exception(exception.args[0]) from None


def fill_element(driver=None, field=None, value=None):
    """
    Fill the given field with the value.
    :param driver: a selenium web driver
    :param field: a dictionary
    :param value: a string or castable to string
    :raise AssertionError: driver is not define, field fails validation
    :raise InvalidElementStateException: if the element is not user-editable
    :raise Exception: web element not found
    :return: 0 if success
    """
    try:
        driver_field_validation(driver, field)
        elem = find_element(driver=driver, field=field)
        if value:
            elem.clear()
            elem.send_keys(str(value))
        else:
            number_of_backspace_hit = len(elem.get_attribute('value'))
            elem.send_keys(number_of_backspace_hit * Keys.BACKSPACE)
        return 0
    except AssertionError as assertion:
        log.error("actions.fill_element raised an assertion with following"
                  " input driver:'{}', field:'{}' and value:'{}'."
                  " Assertion is '{}'".format(driver, field, value, assertion.args))
        raise
    except InvalidElementStateException as invalid_element:
        log.error(invalid_element)
        invalid_element.args = ("Element '{}' must be user-editable".format(field),)
        raise InvalidElementStateException(
            "Element '{}' must be user-editable".format(field)) from None
    except Exception as exception:
        logging.error(
            "actions.fill_element raised an exception. Exception is '{}'".format(
                exception.args))
        raise Exception(
            "actions.fill_element raised an exception. Exception is '{}'".format(
                exception.args[0])) from None


def select_in_dropdown(driver=None, field=None, visible_text=None, value=None):
    """
    see https://stackoverflow.com/questions/7867537/selenium-python-drop-down-menu-option-value
    see https://sqa.stackexchange.com/questions/1355/what-is-the-correct-way-to-select-an-option-using-seleniums-python-webdriver # noqa
    see https://seleniumhq.github.io/selenium/docs/api/py/webdriver_support/selenium.webdriver.support.select.html # noqa
    :raise AssertionError: driver is not define, field is not valid
    :return:
    """
    driver_field_validation(driver, field)

    try:
        my_select = Select(find_element(driver=driver, field=field))

        if visible_text is not None:
            my_select.select_by_visible_text(visible_text)
        elif value is not None:
            my_select.select_by_value(value)
        else:
            raise Exception("select_in_dropdown can't select with no value.")

        return 0

    except NoSuchElementException as no_such_element:
        log.error(no_such_element.args)
        raise NoSuchElementException(no_such_element.args[0]) from None
    except Exception as exception:
        log.error(exception.args)
        raise Exception(exception.args[0]) from None


def select_in_angular_dropdown(driver=None, root_field=None, visible_text: str = None):
    """
    Select a field within a mat_option list

    If root_field is defined then click it and search for visible_text in mat-option elements.
    Otherwise only search for visible_text in mat-options elements
    :param visible_text: a string to search
    :param driver: a selenium web driver
    :param root_field: a dictionary for the container
    :return:
    """
    try:
        if root_field is not None:
            click_element(driver=driver, field=root_field)
            sleep(0.1)

        if is_field_exist(driver=driver, field={"type": "tag_name", "value": "mat-option"}):
            element = find_from_elements(driver=driver,
                                         field={"type": "tag_name", "value": "mat-option"},
                                         text=visible_text)
            element.click()
        else:
            raise Exception('No options displayed within 5 seconds')

    except StaleElementReferenceException as stale_exception:
        log.warning("StaleElementReferenceException. Selection should be done.\n{}".format(
            stale_exception.args))
        return 0
    except Exception as exception:
        log.error(exception.args)
        raise Exception(exception.args[0]) from None


def click_element(driver=None, field=None, web_element: WebElement = None):
    """
    Do simple left click on the given field or on the sub element when web_element is provided
    :param driver: a selenium web driver
    :param field: a dictionary
    :param web_element: an element from which to find the element to click
    :raise AssertionError: driver is not define, field is not valid
    :return: 0 if success
    """
    try:
        driver_field_validation(driver, field)
        web_element_validation(web_element)
        web_element = find_element(driver=driver, field=field, web_element=web_element)
        web_element.click()
        return 0
    except Exception as exception:
        logging.error(exception.args)
        raise Exception(exception.args[0]) from None


def set_checkbox(driver=None, field=None, is_checked=None):
    """
    Set a checkbox to a specific state: True for checked, False for unchecked
    :param driver: a selenium web driver
    :param field: a field identifier as a dictionary
    :param is_checked: the field check value as a boolean (True/False)
    :raise AssertionError: driver is not define, field is not valid, is_checked is not a boolean
    :return: 0 if success
    """
    driver_field_validation(driver, field)
    if not isinstance(is_checked, bool):
        raise AttributeError("is_checked is expected to be a boolean.")

    try:
        elem = find_element(driver=driver, field=field)
        current_state = bool(elem.get_attribute("checked"))
        if current_state is not is_checked:
            elem.click()

        # Check if the element has been set in the expected status
        current_state = bool(elem.get_attribute("checked"))
        if current_state is not is_checked:
            raise Exception(
                "The element '{}' can't be set to the expected status '{}'.".format(field,
                                                                                    is_checked))
        return 0

    except Exception as exception:
        logging.error(exception.args)
        raise Exception(exception.args[0]) from None


def hover_element(driver=None, field=None):
    """
    Do simple hover on the given field
    :param driver: a selenium web driver
    :param field: a dictionary
    :raise AssertionError: driver is not define, field is not valid
    :return: 0 if success
    """
    driver_field_validation(driver, field)
    try:
        elem = find_element(driver=driver, field=field)
        hover = ActionChains(driver).move_to_element(elem)
        hover.perform()
        return 0
    except Exception as exception:
        logging.error(exception.args)
        raise Exception(exception.args[0]) from None


def select_in_elements(driver=None, field=None, displayed_text=None):
    """
    TODO Add the capability to select a field from an webelement child
    Do a click on the element which text
    :param driver: a selenium web driver
    :param field: a dictionary
    :param displayed_text: a string
    :return: 0 if success
    """
    driver_field_validation(driver, field)
    if displayed_text is None or not isinstance(displayed_text, str):
        raise AttributeError("Displayed text must be a non-empty string")
    try:
        if is_field_displayed(driver=driver, field=field):
            element = find_from_elements(driver=driver, field=field, text=displayed_text)
            element.click()
        return 0
    except Exception as exception:
        logging.error(exception.args)
        raise Exception(exception.args[0]) from None
